@@ -22,9 +22,8 @@
 #include "nsCoord.h"
 #include "nsIContentInlines.h"
 #include "nsIScriptError.h"
-#include "nsIStringBundle.h"
 #include "nsContentUtils.h"
-#include "ImageLayers.h"
+#include "nsLayoutUtils.h"
 
 #ifdef ACCESSIBILITY
 #  include "nsAccessibilityService.h"
@@ -74,12 +73,12 @@ inline bool is_space(char c) {
 static void logMessage(nsIContent* aContent, const nsAString& aCoordsSpec,
                        int32_t aFlags, const char* aMessageName) {
   nsContentUtils::ReportToConsole(
-      aFlags, NS_LITERAL_CSTRING("Layout: ImageMap"), aContent->OwnerDoc(),
-      nsContentUtils::eLAYOUT_PROPERTIES, aMessageName, nullptr, /* params */
-      0, /* params length */
+      aFlags, "Layout: ImageMap"_ns, aContent->OwnerDoc(),
+      nsContentUtils::eLAYOUT_PROPERTIES, aMessageName,
+      nsTArray<nsString>(), /* params */
       nullptr,
-      PromiseFlatString(NS_LITERAL_STRING("coords=\"") + aCoordsSpec +
-                        NS_LITERAL_STRING("\""))); /* source line */
+      PromiseFlatString(u"coords=\""_ns + aCoordsSpec +
+                        u"\""_ns)); /* source line */
 }
 
 void Area::ParseCoords(const nsAString& aSpec) {
@@ -468,7 +467,7 @@ void PolyArea::Draw(nsIFrame* aFrame, DrawTarget& aDrawTarget,
       Point p1(pc->CSSPixelsToDevPixels(mCoords[0]),
                pc->CSSPixelsToDevPixels(mCoords[1]));
       Point p2, p1snapped, p2snapped;
-      for (int32_t i = 2; i < mNumCoords; i += 2) {
+      for (int32_t i = 2; i < mNumCoords - 1; i += 2) {
         p2.x = pc->CSSPixelsToDevPixels(mCoords[i]);
         p2.y = pc->CSSPixelsToDevPixels(mCoords[i + 1]);
         p1snapped = p1;
@@ -494,7 +493,7 @@ void PolyArea::GetRect(nsIFrame* aFrame, nsRect& aRect) {
     nscoord x1, x2, y1, y2, xtmp, ytmp;
     x1 = x2 = nsPresContext::CSSPixelsToAppUnits(mCoords[0]);
     y1 = y2 = nsPresContext::CSSPixelsToAppUnits(mCoords[1]);
-    for (int32_t i = 2; i < mNumCoords; i += 2) {
+    for (int32_t i = 2; i < mNumCoords - 1; i += 2) {
       xtmp = nsPresContext::CSSPixelsToAppUnits(mCoords[i]);
       ytmp = nsPresContext::CSSPixelsToAppUnits(mCoords[i + 1]);
       x1 = x1 < xtmp ? x1 : xtmp;
@@ -629,8 +628,8 @@ void nsImageMap::AreaRemoved(HTMLAreaElement* aArea) {
     aArea->SetPrimaryFrame(nullptr);
   }
 
-  aArea->RemoveSystemEventListener(NS_LITERAL_STRING("focus"), this, false);
-  aArea->RemoveSystemEventListener(NS_LITERAL_STRING("blur"), this, false);
+  aArea->RemoveSystemEventListener(u"focus"_ns, this, false);
+  aArea->RemoveSystemEventListener(u"blur"_ns, this, false);
 }
 
 void nsImageMap::FreeAreas() {
@@ -689,7 +688,7 @@ void nsImageMap::UpdateAreas() {
 }
 
 void nsImageMap::AddArea(HTMLAreaElement* aArea) {
-  static Element::AttrValuesArray strings[] = {
+  static AttrArray::AttrValuesArray strings[] = {
       nsGkAtoms::rect,     nsGkAtoms::rectangle,
       nsGkAtoms::circle,   nsGkAtoms::circ,
       nsGkAtoms::_default, nsGkAtoms::poly,
@@ -698,8 +697,8 @@ void nsImageMap::AddArea(HTMLAreaElement* aArea) {
   UniquePtr<Area> area;
   switch (aArea->FindAttrValueIn(kNameSpaceID_None, nsGkAtoms::shape, strings,
                                  eIgnoreCase)) {
-    case Element::ATTR_VALUE_NO_MATCH:
-    case Element::ATTR_MISSING:
+    case AttrArray::ATTR_VALUE_NO_MATCH:
+    case AttrArray::ATTR_MISSING:
     case 0:
     case 1:
       area = MakeUnique<RectArea>(aArea);
@@ -722,8 +721,8 @@ void nsImageMap::AddArea(HTMLAreaElement* aArea) {
   }
 
   // Add focus listener to track area focus changes
-  aArea->AddSystemEventListener(NS_LITERAL_STRING("focus"), this, false, false);
-  aArea->AddSystemEventListener(NS_LITERAL_STRING("blur"), this, false, false);
+  aArea->AddSystemEventListener(u"focus"_ns, this, false, false);
+  aArea->AddSystemEventListener(u"blur"_ns, this, false, false);
 
   // This is a nasty hack.  It needs to go away: see bug 135040.  Once this is
   // removed, the code added to RestyleManager::RestyleElement,
@@ -738,10 +737,10 @@ void nsImageMap::AddArea(HTMLAreaElement* aArea) {
   mAreas.AppendElement(std::move(area));
 }
 
-HTMLAreaElement* nsImageMap::GetArea(nscoord aX, nscoord aY) const {
+HTMLAreaElement* nsImageMap::GetArea(const CSSIntPoint& aPt) const {
   NS_ASSERTION(mMap, "Not initialized");
   for (const auto& area : mAreas) {
-    if (area->IsInside(aX, aY)) {
+    if (area->IsInside(aPt.x, aPt.y)) {
       return area->mArea;
     }
   }
