@@ -13,14 +13,13 @@
 namespace mozilla {
 
 nscoord ServoStyleSet::EvaluateSourceSizeList(
-    const RawServoSourceSizeList* aSourceSizeList) const {
-  return Servo_SourceSizeList_Evaluate(mRawSet.get(), aSourceSizeList);
+    const StyleSourceSizeList* aSourceSizeList) const {
+  return Servo_SourceSizeList_Evaluate(mRawData.get(), aSourceSizeList);
 }
 
 already_AddRefed<ComputedStyle> ServoStyleSet::ResolveServoStyle(
     const dom::Element& aElement) {
-  // TODO(emilio): mRawSet isn't used by this function, remove.
-  return Servo_ResolveStyle(&aElement, mRawSet.get()).Consume();
+  return Servo_ResolveStyle(&aElement).Consume();
 }
 
 }  // namespace mozilla
