@@ -5,17 +5,18 @@
 //! Style resolution for a given element or pseudo-element.
 
 use crate::applicable_declarations::ApplicableDeclarationList;
+use crate::computed_value_flags::ComputedValueFlags;
 use crate::context::{CascadeInputs, ElementCascadeInputs, StyleContext};
 use crate::data::{EagerPseudoStyles, ElementStyles};
 use crate::dom::TElement;
 use crate::matching::MatchMethods;
 use crate::properties::longhands::display::computed_value::T as Display;
-use crate::properties::{AnimationRules, ComputedValues};
+use crate::properties::ComputedValues;
 use crate::rule_tree::StrongRuleNode;
 use crate::selector_parser::{PseudoElement, SelectorImpl};
 use crate::stylist::RuleInclusion;
 use log::Level::Trace;
-use selectors::matching::{ElementSelectorFlags, MatchingContext};
+use selectors::matching::{MatchingContext, NeedsSelectorFlags};
 use selectors::matching::{MatchingMode, VisitedHandlingMode};
 use servo_arc::Arc;
 
@@ -44,6 +45,7 @@ where
 
 struct MatchingResults {
     rule_node: StrongRuleNode,
+    flags: ComputedValueFlags,
 }
 
 /// A style returned from the resolver machinery.
@@ -66,6 +68,18 @@ pub struct ResolvedElementStyles {
     pub pseudos: EagerPseudoStyles,
 }
 
+impl ResolvedElementStyles {
+    /// Convenience accessor for the primary style.
+    pub fn primary_style(&self) -> &Arc<ComputedValues> {
+        &self.primary.style.0
+    }
+
+    /// Convenience mutable accessor for the style.
+    pub fn primary_style_mut(&mut self) -> &mut Arc<ComputedValues> {
+        &mut self.primary.style.0
+    }
+}
+
 impl PrimaryStyle {
     /// Convenience accessor for the style.
     pub fn style(&self) -> &ComputedValues {
@@ -106,23 +120,36 @@ where
     )
 }
 
+fn layout_parent_style_for_pseudo<'a>(
+    primary_style: &'a PrimaryStyle,
+    layout_parent_style: Option<&'a ComputedValues>,
+) -> Option<&'a ComputedValues> {
+    if primary_style.style().is_display_contents() {
+        layout_parent_style
+    } else {
+        Some(primary_style.style())
+    }
+}
+
 fn eager_pseudo_is_definitely_not_generated(
     pseudo: &PseudoElement,
     style: &ComputedValues,
 ) -> bool {
-    use crate::properties::computed_value_flags::ComputedValueFlags;
-
     if !pseudo.is_before_or_after() {
         return false;
     }
 
-    if !style.flags.intersects(ComputedValueFlags::INHERITS_DISPLAY) &&
+    if !style
+        .flags
+        .intersects(ComputedValueFlags::DISPLAY_DEPENDS_ON_INHERITED_STYLE) &&
         style.get_box().clone_display() == Display::None
     {
         return true;
     }
 
-    if !style.flags.intersects(ComputedValueFlags::INHERITS_CONTENT) &&
+    if !style
+        .flags
+        .intersects(ComputedValueFlags::CONTENT_DEPENDS_ON_INHERITED_STYLE) &&
         style.ineffective_content_property()
     {
         return true;
@@ -177,6 +204,7 @@ where
             CascadeInputs {
                 rules: Some(primary_results.rule_node),
                 visited_rules,
+                flags: primary_results.flags,
             },
             parent_style,
             layout_parent_style,
@@ -191,7 +219,7 @@ where
     ) -> PrimaryStyle {
         // Before doing the cascade, check the sharing cache and see if we can
         // reuse the style via rule node identity.
-        let may_reuse = !self.element.is_in_native_anonymous_subtree() &&
+        let may_reuse = self.element.matches_user_and_content_rules() &&
             parent_style.is_some() &&
             inputs.rules.is_some();
 
@@ -233,12 +261,9 @@ where
 
         let mut pseudo_styles = EagerPseudoStyles::default();
 
-        if self.element.implemented_pseudo_element().is_none() {
-            let layout_parent_style_for_pseudo = if primary_style.style().is_display_contents() {
-                layout_parent_style
-            } else {
-                Some(primary_style.style())
-            };
+        if !self.element.is_pseudo_element() {
+            let layout_parent_style_for_pseudo =
+                layout_parent_style_for_pseudo(&primary_style, layout_parent_style);
             SelectorImpl::each_eagerly_cascaded_pseudo_element(|pseudo| {
                 let pseudo_style = self.resolve_pseudo_style(
                     &pseudo,
@@ -286,6 +311,26 @@ where
         })
     }
 
+    /// Cascade a set of rules for pseudo element, using the default parent for inheritance.
+    pub fn cascade_style_and_visited_for_pseudo_with_default_parents(
+        &mut self,
+        inputs: CascadeInputs,
+        pseudo: &PseudoElement,
+        primary_style: &PrimaryStyle,
+    ) -> ResolvedStyle {
+        with_default_parent_styles(self.element, |_, layout_parent_style| {
+            let layout_parent_style_for_pseudo =
+                layout_parent_style_for_pseudo(primary_style, layout_parent_style);
+
+            self.cascade_style_and_visited(
+                inputs,
+                Some(primary_style.style()),
+                layout_parent_style_for_pseudo,
+                Some(pseudo),
+            )
+        })
+    }
+
     fn cascade_style_and_visited(
         &mut self,
         inputs: CascadeInputs,
@@ -293,10 +338,6 @@ where
         layout_parent_style: Option<&ComputedValues>,
         pseudo: Option<&PseudoElement>,
     ) -> ResolvedStyle {
-        debug_assert!(
-            self.element.implemented_pseudo_element().is_none() || pseudo.is_none(),
-            "Pseudo-elements can't have other pseudos!"
-        );
         debug_assert!(pseudo.map_or(true, |p| p.is_eager()));
 
         let implemented_pseudo = self.element.implemented_pseudo_element();
@@ -308,10 +349,10 @@ where
             pseudo,
             inputs,
             &self.context.shared.guards,
+            pseudo.and(parent_style),
             parent_style,
             parent_style,
             layout_parent_style,
-            &self.context.thread_local.font_metrics_provider,
             Some(&self.context.thread_local.rule_cache),
             &mut conditions,
         );
@@ -379,25 +420,34 @@ where
         originating_element_style: &PrimaryStyle,
         layout_parent_style: Option<&ComputedValues>,
     ) -> Option<ResolvedStyle> {
-        let rules = self.match_pseudo(
-            originating_element_style.style(),
+        let MatchingResults {
+            rule_node,
+            mut flags,
+        } = self.match_pseudo(
+            &originating_element_style.style.0,
             pseudo,
             VisitedHandlingMode::AllLinksUnvisited,
         )?;
 
         let mut visited_rules = None;
         if originating_element_style.style().visited_style().is_some() {
-            visited_rules = self.match_pseudo(
-                originating_element_style.style(),
-                pseudo,
-                VisitedHandlingMode::RelevantLinkVisited,
-            );
+            visited_rules = self
+                .match_pseudo(
+                    &originating_element_style.style.0,
+                    pseudo,
+                    VisitedHandlingMode::RelevantLinkVisited,
+                )
+                .map(|results| {
+                    flags |= results.flags;
+                    results.rule_node
+                });
         }
 
         Some(self.cascade_style_and_visited(
             CascadeInputs {
-                rules: Some(rules),
+                rules: Some(rule_node),
                 visited_rules,
+                flags,
             },
             Some(originating_element_style.style()),
             layout_parent_style,
@@ -412,38 +462,30 @@ where
         );
         let mut applicable_declarations = ApplicableDeclarationList::new();
 
-        let map = &mut self.context.thread_local.selector_flags;
         let bloom_filter = self.context.thread_local.bloom_filter.filter();
         let nth_index_cache = &mut self.context.thread_local.nth_index_cache;
         let mut matching_context = MatchingContext::new_for_visited(
             MatchingMode::Normal,
             Some(bloom_filter),
-            Some(nth_index_cache),
+            nth_index_cache,
             visited_handling,
             self.context.shared.quirks_mode(),
+            NeedsSelectorFlags::Yes,
         );
 
         let stylist = &self.context.shared.stylist;
         let implemented_pseudo = self.element.implemented_pseudo_element();
-        {
-            let resolving_element = self.element;
-            let mut set_selector_flags = |element: &E, flags: ElementSelectorFlags| {
-                resolving_element.apply_selector_flags(map, element, flags);
-            };
-
-            // Compute the primary rule node.
-            stylist.push_applicable_declarations(
-                self.element,
-                implemented_pseudo.as_ref(),
-                self.element.style_attribute(),
-                self.element.smil_override(),
-                self.element.animation_rules(),
-                self.rule_inclusion,
-                &mut applicable_declarations,
-                &mut matching_context,
-                &mut set_selector_flags,
-            );
-        }
+        // Compute the primary rule node.
+        stylist.push_applicable_declarations(
+            self.element,
+            implemented_pseudo.as_ref(),
+            self.element.style_attribute(),
+            self.element.smil_override(),
+            self.element.animation_declarations(self.context.shared),
+            self.rule_inclusion,
+            &mut applicable_declarations,
+            &mut matching_context,
+        );
 
         // FIXME(emilio): This is a hack for animations, and should go away.
         self.element.unset_dirty_style_attribute();
@@ -462,7 +504,20 @@ where
             }
         }
 
-        MatchingResults { rule_node }
+        if matching_context.considered_relative_selector {
+            // This is a bit awkward - ideally, the flag is set directly where `considered_relative_selector`
+            // is; however, in that context, the implementation detail of `extra_data` is not visible, so
+            // it's done here. A trait for manipulating the flags is an option, but not worth it for a single flag.
+            matching_context
+                .extra_data
+                .cascade_input_flags
+                .insert(ComputedValueFlags::CONSIDERED_RELATIVE_SELECTOR);
+        }
+
+        MatchingResults {
+            rule_node,
+            flags: matching_context.extra_data.cascade_input_flags,
+        }
     }
 
     fn match_pseudo(
@@ -470,15 +525,15 @@ where
         originating_element_style: &ComputedValues,
         pseudo_element: &PseudoElement,
         visited_handling: VisitedHandlingMode,
-    ) -> Option<StrongRuleNode> {
+    ) -> Option<MatchingResults> {
         debug!(
             "Match pseudo {:?} for {:?}, visited: {:?}",
             self.element, pseudo_element, visited_handling
         );
         debug_assert!(pseudo_element.is_eager());
         debug_assert!(
-            self.element.implemented_pseudo_element().is_none(),
-            "Element pseudos can't have any other pseudo."
+            !self.element.is_pseudo_element(),
+            "Element pseudos can't have any other eager pseudo."
         );
 
         let mut applicable_declarations = ApplicableDeclarationList::new();
@@ -495,19 +550,15 @@ where
         let bloom_filter = self.context.thread_local.bloom_filter.filter();
         let nth_index_cache = &mut self.context.thread_local.nth_index_cache;
 
-        let mut matching_context = MatchingContext::new_for_visited(
+        let mut matching_context = MatchingContext::<'_, E::Impl>::new_for_visited(
             MatchingMode::ForStatelessPseudoElement,
             Some(bloom_filter),
-            Some(nth_index_cache),
+            nth_index_cache,
             visited_handling,
             self.context.shared.quirks_mode(),
+            NeedsSelectorFlags::Yes,
         );
-
-        let map = &mut self.context.thread_local.selector_flags;
-        let resolving_element = self.element;
-        let mut set_selector_flags = |element: &E, flags: ElementSelectorFlags| {
-            resolving_element.apply_selector_flags(map, element, flags);
-        };
+        matching_context.extra_data.originating_element_style = Some(originating_element_style);
 
         // NB: We handle animation rules for ::before and ::after when
         // traversing them.
@@ -516,11 +567,10 @@ where
             Some(pseudo_element),
             None,
             None,
-            AnimationRules(None, None),
+            /* animation_declarations = */ Default::default(),
             self.rule_inclusion,
             &mut applicable_declarations,
             &mut matching_context,
-            &mut set_selector_flags,
         );
 
         if applicable_declarations.is_empty() {
@@ -531,6 +581,9 @@ where
             .rule_tree()
             .compute_rule_node(&mut applicable_declarations, &self.context.shared.guards);
 
-        Some(rule_node)
+        Some(MatchingResults {
+            rule_node,
+            flags: matching_context.extra_data.cascade_input_flags,
+        })
     }
 }
